"""
====================
rororo.schemas.empty
====================

Collection of empty JSON Schemas, useful for validation empty requests or
responses.

"""

EMPTY_ARRAY = {
    'type': 'array',
    'maxItems': 0,
}
EMPTY_OBJECT = {
    'type': 'object',
    'additionalProperties': False,
}



"""
===============
rororo.settings
===============

Immutable Settings dictionary and various utilities to read settings values
from environment.

Module helps you to prepare and read settings inside your web application.

"""

import calendar
import os
import time
import types

from distutils.util import strtobool
from importlib import import_module
from locale import LC_ALL, setlocale
from logging.config import dictConfig as setup_logging  # noqa


def from_env(key, default=None):
    """Shortcut for safely reading environment variable.

    :param key: Environment var key.
    :param default:
        Return default value if environment var not found by given key. By
        default: None
    """
    return os.environ.get(key, default)


def immutable_settings(defaults, **optionals):
    r"""Initialize and return immutable Settings dictionary.

    Settings dictionary allows you to setup settings values from multiple
    sources and make sure that values cannot be changed, updated by anyone else
    after initialization. This helps keep things clear and not worry about
    hidden settings change somewhere around your web application.

    :param defaults:
       Read settings values from module or dict-like instance.
    :type defaults: module or dict
    :param \*\*optionals:
        Update base settings with optional values.

        In common additional values shouldn't be passed, if settings values
        already populated from local settings or environment. But in case
        of using application factories this makes sense::

            from . import settings

            def create_app(**options):
                app = ...
                app.settings = Settings(settings, **options)
                return app

        And yes each additional key overwrite default setting value.
    :type /*/*optionals: dict
    """
    settings = {key: value for key, value in iter_settings(defaults)}
    for key, value in iter_settings(optionals):
        settings[key] = value
    return types.MappingProxyType(settings)


def inject_settings(mixed, context, fail_silently=False):
    """Inject settings values to given context.

    :param mixed:
        Settings could read from Python path, Python module or dict-like
        instance.
    :type mixed: str, module or dict
    :param context: Context should support dict-like item assingment.
    :type context: dict
    :param fail_silently:
        When enabled and reading settings from Python path ignore errors if
        given Python path couldn't be loaded.
    """
    if isinstance(mixed, str):
        try:
            mixed = import_module(mixed)
        except Exception:
            if fail_silently:
                return
            raise

    for key, value in iter_settings(mixed):
        context[key] = value


def is_setting_key(key):
    """Check whether given key is valid setting key or not.

    Only public uppercase constants are valid settings keys, all other keys
    are invalid and shouldn't present in Settings dict.

    Valid settings keys
    ===================

    ::

        DEBUG
        SECRET_KEY

    Invalid settings keys
    =====================

    ::

        _PRIVATE_SECRET_KEY
        camelCasedSetting
        rel
        secret_key

    :param key: Key to check.
    :type key: str
    """
    return key.isupper() and key[0] != '_'


def iter_settings(mixed):
    """Iterate over settings values from settings module or dict-like instance.

    :param mixed: Settings instance to iterate.
    :type mixed: module or dict
    """
    if isinstance(mixed, types.ModuleType):
        for attr in dir(mixed):
            if not is_setting_key(attr):
                continue
            yield (attr, getattr(mixed, attr))
    else:
        yield from filter(lambda item: is_setting_key(item[0]), mixed.items())


def setup_locale(locale, first_weekday=None):
    """Setup locale for backend application.

    :param locale: Locale to use.
    :param first_weekday:
        Weekday for start week. 0 for Monday, 6 for Sunday. By default: None
    """
    if first_weekday is not None:
        calendar.setfirstweekday(first_weekday)
    return setlocale(LC_ALL, locale)


def setup_timezone(timezone):
    """Setup timezone for backend application.

    :param timezone: Timezone to use, e.g. "UTC", "Europe/Kiev".
    :type timezone: str
    """
    if timezone and hasattr(time, 'tzset'):
        tz_root = '/usr/share/zoneinfo'
        tz_filename = os.path.join(tz_root, *(timezone.split('/')))

        if os.path.exists(tz_root) and not os.path.exists(tz_filename):
            raise ValueError('Incorrect timezone value: {0}'.format(timezone))

        os.environ['TZ'] = timezone
        time.tzset()


def to_bool(value):
    """Convert string or other Python value to boolean.

    Rationalle
    ==========

    Passing flags is one of the most used cases of using environment vars and
    as values are strings we need easy way to convert them to something
    suitable with Python for later usage.

    And when ``int`` or ``float`` values should be converted without extra
    work, ``bool('0') => True``. So in case of expecting boolean flag from
    environment just use this function for all good things.

    :param value: String or other value.
    """
    return bool(strtobool(value) if isinstance(value, str) else value)



"""
=============
rororo.logger
=============

Logging utilities.

Module provides easy way to setup logging for your web application.

"""

import sys


class IgnoreErrorsFilter(object):

    """Ignore all warnings and errors from stdout handler."""

    def filter(self, record):
        """Allow only debug and info log messages to stdout handler."""
        return record.levelname in {'DEBUG', 'INFO'}


def default_logging_dict(*loggers, **kwargs):
    r"""Prepare logging dict suitable with ``logging.config.dictConfig``.

    :param \*loggers: Enable logging for each logger in sequence.
    :param \*\*kwargs: Setup additional logger params via keyword arguments.
    """
    kwargs.setdefault('level', 'INFO')
    return {
        'version': 1,
        'disable_existing_loggers': True,
        'filters': {
            'ignore_errors': {
                '()': IgnoreErrorsFilter,
            },
        },
        'formatters': {
            'default': {
                'format': '%(asctime)s [%(levelname)s:%(name)s] %(message)s',
            },
            'naked': {
                'format': u'%(message)s',
            },
        },
        'handlers': {
            'stdout': {
                'class': 'logging.StreamHandler',
                'filters': ['ignore_errors'],
                'formatter': 'default',
                'level': 'DEBUG',
                'stream': sys.stdout,
            },
            'stderr': {
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                'level': 'WARNING',
                'stream': sys.stderr,
            },
        },
        'loggers': {
            logger: dict(handlers=['stdout', 'stderr'], **kwargs)
            for logger in loggers
        },
    }


def update_sentry_logging(logging_dict, sentry_dsn, *loggers):
    r"""Enable Sentry logging if Sentry DSN passed.

    .. note::
        Sentry logging requires `raven <http://pypi.python.org/pypi/raven>`_
        library to be installed.

    :param logging_dict: Logging dict.
    :param sentry_dsn:
        Sentry DSN value. If ``None`` do not update logging dict at all.
    :param \*loggers:
        Use Sentry logging for each logger in the sequence. If the sequence is
        empty use Sentry logging to each available logger.
    """
    # No Sentry DSN, nothing to do
    if not sentry_dsn:
        return

    # Add Sentry handler
    logging_dict['handlers']['sentry'] = {
        'level': 'WARNING',
        'class': 'raven.handlers.logging.SentryHandler',
        'dsn': sentry_dsn,
    }

    loggers = tuple(logging_dict['loggers']) if not loggers else loggers
    for logger in loggers:
        # Ignore missing loggers
        logger_dict = logging_dict['loggers'].get(logger)
        if not logger_dict:
            continue

        # Ignore logger from logger config
        if logger_dict.pop('ignore_sentry', False):
            continue

        # Handlers list should exist
        handlers = list(logger_dict.setdefault('handlers', []))
        handlers.append('sentry')
        logger_dict['handlers'] = tuple(handlers)



"""
======
rororo
======

Collection of utilities, helpers, and principles for building Python backend
applications. Supports `aiohttp.web <http://aiohttp.readthedocs.org/>`_,
`Flask <http://flask.pocoo.org/>`_, and your web-framework.

"""

__author__ = 'Igor Davydenko'
__license__ = 'BSD'
__version__ = '1.0.0a1'



"""
=====================
rororo.schemas.schema
=====================

Implement class for validating request and response data against JSON Schema.

"""

import logging
import types

try:
    from aiohttp.multidict import MultiDict, MultiDictProxy
except ImportError:  # pragma: no cover
    MultiDict, MultiDictProxy = None, None

from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate

from .exceptions import Error
from .utils import defaults
from .validators import Validator


logger = logging.getLogger(__name__)


class Schema(object):

    """Validate request and response data against JSON Schema."""

    __slots__ = ('error_class', 'module', 'response_factory', 'validator',
                 '_valid_request')

    def __init__(self, module, *, response_factory=None, error_class=None,
                 validator=None):
        """Initialize Schema object.

        :param module: Module contains at least request and response schemas.
        :param response_factory: Put valid response data to this factory func.
        :param error_class:
            Wrap all errors in given class. If empty real errors would be
            reraised.
        :param validator: Use given validator instead of standart one.
        """
        self._valid_request = None
        self.error_class = error_class
        self.module = module
        self.response_factory = response_factory
        self.validator = validator or Validator

    def make_error(self, message, *, error=None):
        """Return error instantiated from given message.

        :param message: Message to wrap.
        :param error: Validation error.
        """
        return (Error(message)
                if self.error_class is None
                else self.error_class(message))

    def make_response(self, data=None, **kwargs):
        r"""Validate response data and wrap it inside response factory.

        :param data: Response data. Could be ommited.
        :param \*\*kwargs: Keyword arguments to be passed to response factory.
        """
        if not self._valid_request:
            logger.error('Request not validated, cannot make response')
            raise self.make_error('Request not validated before, cannot make '
                                  'response')

        if data is None and self.response_factory is None:
            logger.error('Response data omit, but no response factory is used')
            raise self.make_error('Response data could be omitted only when '
                                  'response factory is used')

        response_schema = getattr(self.module, 'response', None)
        if response_schema is not None:
            self._validate(data, response_schema)

        if self.response_factory is not None:
            args = (data, ) if data is not None else ()
            return self.response_factory(*args, **kwargs)
        return data

    def validate_request(self, data, *additional, merged_class=dict):
        r"""Validate request data against request schema from module.

        :param data: Request data.
        :param \*additional:
            Additional data dicts to be merged with base request data.
        :param merged_class:
            When additional data dicts supplied method by default will return
            merged **dict** with all data, but you can customize things to
            use read-only dict or any other additional class or callable.
        """
        request_schema = getattr(self.module, 'request', None)
        if request_schema is None:
            logger.error('Request schema should be defined',
                         extra={'schema_module': self.module,
                                'schema_module_attrs': dir(self.module)})
            raise self.make_error('Request schema should be defined')

        # Merge base and additional data dicts, but only if additional data
        # dicts have been supplied
        if additional:
            data = merged_class(self._merge_data(data, *additional))

        try:
            self._validate(data, request_schema)
        finally:
            self._valid_request = False

        self._valid_request = True

        processor = getattr(self.module, 'request_processor', None)
        return processor(data) if processor else data

    def _merge_data(self, data, *additional):
        r"""Merge base data and additional dicts.

        :param data: Base data.
        :param \*additional: Additional data dicts to be merged into base dict.
        """
        data = self._pure_data(data)
        for item in additional:
            data = defaults(data, self._pure_data(item))
        return data

    def _pure_data(self, data):
        """
        Convert multi-dicts or any other custom dicts to normal dicts to be
        compatible with ``jsonschema.validate`` function.

        :param data: Python dict or other object.
        """
        dict_types = [types.MappingProxyType]
        if MultiDict is not None:  # pragma: no branch
            dict_types.extend((MultiDict, MultiDictProxy))

        if isinstance(data, tuple(dict_types)):
            return dict(data)

        return data

    def _validate(self, data, schema):
        """Validate data against given schema.

        :param data: Data to validate.
        :param schema: Schema to use for validation.
        """
        try:
            return validate(self._pure_data(data), schema, self.validator)
        except ValidationError as err:
            logger.error('Schema validation error',
                         exc_info=True,
                         extra={'schema': schema,
                                'schema_module': self.module})
            if self.error_class is None:
                raise
            raise self.make_error('Validation Error', error=err) from err



"""
=========================
rororo.schemas.exceptions
=========================

Exceptions for Schemas.

.. note:: If data cannot be validate against given schema ValidationError from
   ``jsonschema`` library would be raised, not ``rororo.schemas.Error``.

"""


class Error(Exception):

    """Base Schema exception."""



"""
==============
rororo.schemas
==============

Validate request and response data using JSON Schema.

"""

from .exceptions import Error  # noqa
from .schema import Schema  # noqa
from .validators import Validator  # noqa
from .utils import defaults  # noqa



#!/usr/bin/env python

import os
import re

from pathlib import Path
from setuptools import setup


rel = Path(__file__).parent


with open(str(rel / 'README.rst')) as handler:
    README = handler.read()

with open(str(rel / 'rororo' / '__init__.py')) as handler:
    INIT_PY = handler.read()

VERSION = re.findall("__version__ = '([^']+)'", INIT_PY)[0]


setup(
    name='rororo',
    version=VERSION,
    description=(
        'Collection of utilities, helpers, and principles for building Python '
        'backend applications. Supports aiohttp.web, Flask, and your '
        'web-framework'
    ),
    long_description=README,
    author='Igor Davydenko',
    author_email='playpauseandstop@gmail.com',
    url='http://github.com/playpauseandstop/rororo',
    packages=[
        'rororo',
        'rororo.schemas',
    ],
    install_requires=[
        'jsonschema>=2.4.0',
    ],
    platforms='any',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Utilities',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'License :: OSI Approved :: BSD License',
    ],
    keywords='utilities helpers principles aiohttp flask web framework',
    license='BSD License',
)



"""
====================
rororo.schemas.utils
====================

Utilities for Schema package.

"""


def defaults(current, *args):
    r"""Override current dict with defaults values.

    :param current: Current dict.
    :param \*args: Sequence with default data dicts.
    """
    for data in args:
        for key, value in data.items():
            current.setdefault(key, value)
    return current



"""
=========================
rororo.schemas.validators
=========================

Customize default JSON Schema Draft 4 validator.

"""

from jsonschema.validators import Draft4Validator

from .utils import defaults


class Validator(Draft4Validator):

    """Customize default JSON Schema validator.

    This customization allows:

    * Use tuples for "array" data type

    """

    DEFAULT_TYPES = defaults({
        'array': (list, tuple),
    }, Draft4Validator.DEFAULT_TYPES)
